/*
 * Test dataspace operations.
 *
 * Copyright (C) 2020, 2021 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/re/env.h>
#include <l4/sys/err.h>

#include <systypes/fcntl.h>

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "file.h"
#include "memory_utils.h"



static void show(file_t *file, unsigned long step, unsigned long sample)
{
  /* Allocate a buffer for sampling from the file. */

  char buf[sample + 1];

  for (unsigned long offset = 0; offset < file_populated_span(file); offset += step)
  {
    unsigned long remaining = file_populated_span(file) - offset;
    unsigned long sample_remaining = remaining < sample ? remaining : sample;

    printf("%ld bytes from %p...\n", sample_remaining, (file->memory + offset));
    strncpy(buf, (file->memory + offset), sample_remaining);
    buf[sample_remaining] = '\0';
    printf("%s\n", buf);
  }
}

int main(int argc, char *argv[])
{
  if (argc < 4)
  {
    printf("Need filename, step and sample size.\n");
    return 1;
  }

  /* Obtain filename and access parameters. */

  char *filename = argv[1];
  unsigned long step = atoi(argv[2]);
  unsigned long sample = atoi(argv[3]);

  /* Obtain access to the filesystem. */

  l4_cap_idx_t server = l4re_env_get_cap("server");

  /* Invoke the open method to receive the file reference. */

  file_t file;
  long err = file_open(&file, filename, O_RDWR, server);

  if (err)
  {
    printf("Could not obtain file: %s\n", l4sys_errtostr(err));
    return 1;
  }

  /* A region of the file is mapped. */

  err = file_mmap(&file, 0, page(10));

  if (err)
  {
    printf("Could not map file region: %s\n", l4sys_errtostr(err));
    return 1;
  }

  show(&file, step, sample);

  /* Resizing must occur before writing beyond the end of file. Otherwise, the
     data may get discarded if the supporting flexpage needs to be flushed. */

  offset_t new_region = round(file_populated_span(&file), page(1));

  printf("Resize to %ld...\n", new_region + file_populated_span(&file));

  err = file_resize(&file, new_region + file_populated_span(&file));

  if (err)
  {
    printf("Could not resize file: %s\n", l4sys_errtostr(err));
    return 1;
  }

  printf("Resized file...\n");

  /* Copy the sampled data to another file region. */

  printf("Copy data to %ld...\n", new_region);

  for (unsigned long offset = 0; offset < file_populated_span(&file); offset += step)
  {
    memcpy(file.memory + new_region + offset, file.memory + offset, sample);
    if (step > sample)
      memset(file.memory + new_region + offset + sample, 0, step - sample);
  }

  show(&file, step, sample);

  printf("File shown.\n");

  return 0;
}

// vim: tabstop=2 expandtab shiftwidth=2
